@@ -8,30 +8,50 @@ SCENARIO_DIR = osp.dirname(osp.abspath(__file__))
 # default subnet address for internet
 INTERNET = 0
 
+# Constants
+NUM_ACCESS_LEVELS = 2
+NO_ACCESS = 0
+USER_ACCESS = 1
+ROOT_ACCESS = 2
+
 # scenario property keys
 SUBNETS = "subnets"
 TOPOLOGY = "topology"
 SENSITIVE_HOSTS = "sensitive_hosts"
 SERVICES = "services"
 OS = "os"
+PROCESSES = "processes"
 EXPLOITS = "exploits"
+PRIVESCS = "privilege_escalation"
 SERVICE_SCAN_COST = "service_scan_cost"
 OS_SCAN_COST = "os_scan_cost"
 SUBNET_SCAN_COST = "subnet_scan_cost"
+PROCESS_SCAN_COST = "process_scan_cost"
 HOST_CONFIGS = "host_configurations"
 FIREWALL = "firewall"
 HOSTS = "host"
 STEP_LIMIT = "step_limit"
+ACCESS_LEVELS = "access_levels"
 
 # scenario exploit keys
 EXPLOIT_SERVICE = "service"
 EXPLOIT_OS = "os"
 EXPLOIT_PROB = "prob"
 EXPLOIT_COST = "cost"
+EXPLOIT_ACCESS = "access"
+
+# scenario privilege escalation keys
+PRIVESC_PROCESS = "process"
+PRIVESC_OS = "os"
+PRIVESC_PROB = "prob"
+PRIVESC_COST = "cost"
+PRIVESC_ACCESS = "access"
 
 # host configuration keys
 HOST_SERVICES = "services"
+HOST_PROCESSES = "processes"
 HOST_OS = "os"
+HOST_FIREWALL = "firewall"
 
 
 def load_yaml(file_path):
